@@ -1,32 +1,5 @@
-///////////////////////////////////////////////////////////////////////////
-//
-// Copyright (c) 2012-2018 DreamWorks Animation LLC
-//
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
-//
-// Redistributions of source code must retain the above copyright
-// and license notice and the following restrictions and disclaimer.
-//
-// *     Neither the name of DreamWorks Animation nor the names of
-// its contributors may be used to endorse or promote products derived
-// from this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-// IN NO EVENT SHALL THE COPYRIGHT HOLDERS' AND CONTRIBUTORS' AGGREGATE
-// LIABILITY FOR ALL CLAIMS REGARDLESS OF THEIR BASIS EXCEED US$250.00.
-//
-///////////////////////////////////////////////////////////////////////////
+// Copyright Contributors to the OpenVDB Project
+// SPDX-License-Identifier: MPL-2.0
 
 /// @file points/AttributeArray.h
 ///
@@ -48,9 +21,10 @@
 #include "StreamCompression.h"
 
 #include <tbb/spin_mutex.h>
-#include <tbb/atomic.h>
+#include <atomic>
 
 #include <memory>
+#include <mutex>
 #include <string>
 #include <type_traits>
 
@@ -78,7 +52,7 @@ floatingPointToFixedPoint(const FloatT s)
     static_assert(std::is_unsigned<IntegerT>::value, "IntegerT must be unsigned");
     if (FloatT(0.0) > s) return std::numeric_limits<IntegerT>::min();
     else if (FloatT(1.0) <= s) return std::numeric_limits<IntegerT>::max();
-    return IntegerT(std::floor(s * FloatT(std::numeric_limits<IntegerT>::max())));
+    return IntegerT(s * FloatT(std::numeric_limits<IntegerT>::max()));
 }
 
 
@@ -127,7 +101,6 @@ public:
     enum Flag {
         TRANSIENT = 0x1,            /// by default not written to disk
         HIDDEN = 0x2,               /// hidden from UIs or iterators
-        OUTOFCORE = 0x4,            /// data not yet loaded from disk (deprecated flag as of ABI=5)
         CONSTANTSTRIDE = 0x8,       /// stride size does not vary in the array
         STREAMING = 0x10,           /// streaming mode collapses attributes when first accessed
         PARTIALREAD = 0x20          /// data has been partially read (compressed bytes is used)
@@ -141,56 +114,40 @@ public:
         WRITEPAGED = 0x8            /// data is written out in pages
     };
 
+    // Scoped Lock wrapper class that locks the AttributeArray registry mutex
+    class OPENVDB_API ScopedRegistryLock
+    {
+        tbb::spin_mutex::scoped_lock lock;
+    public:
+        ScopedRegistryLock();
+    }; // class ScopedRegistryLock
+
     using Ptr           = std::shared_ptr<AttributeArray>;
     using ConstPtr      = std::shared_ptr<const AttributeArray>;
 
-    using FactoryMethod = Ptr (*)(Index, Index, bool);
+    using FactoryMethod = Ptr (*)(Index, Index, bool, const Metadata*);
 
     template <typename ValueType, typename CodecType> friend class AttributeHandle;
 
-    AttributeArray() : mPageHandle() { }
+    AttributeArray(): mPageHandle() { mOutOfCore = 0; }
     virtual ~AttributeArray()
     {
         // if this AttributeArray has been partially read, zero the compressed bytes,
         // so the page handle won't attempt to clean up invalid memory
         if (mFlags & PARTIALREAD)       mCompressedBytes = 0;
     }
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
-    AttributeArray(const AttributeArray& rhs)
-        : mIsUniform(rhs.mIsUniform)
-        , mFlags(rhs.mFlags)
-        , mSerializationFlags(rhs.mSerializationFlags)
-        , mOutOfCore(rhs.mOutOfCore)
-        , mPageHandle()
-    {
-        if (mFlags & PARTIALREAD)       mCompressedBytes = rhs.mCompressedBytes;
-        else if (rhs.mPageHandle)       mPageHandle = rhs.mPageHandle->copy();
-    }
-    AttributeArray& operator=(const AttributeArray& rhs)
-    {
-        // if this AttributeArray has been partially read, zero the compressed bytes,
-        // so the page handle won't attempt to clean up invalid memory
-        if (mFlags & PARTIALREAD)       mCompressedBytes = 0;
-        mIsUniform = rhs.mIsUniform;
-        mFlags = rhs.mFlags;
-        mSerializationFlags = rhs.mSerializationFlags;
-        mOutOfCore = rhs.mOutOfCore;
-        if (mFlags & PARTIALREAD)       mCompressedBytes = rhs.mCompressedBytes;
-        else if (rhs.mPageHandle)       mPageHandle = rhs.mPageHandle->copy();
-        else                            mPageHandle.reset();
-        return *this;
-    }
-#else
-    AttributeArray(const AttributeArray&) = default;
-    AttributeArray& operator=(const AttributeArray&) = default;
-#endif
-    AttributeArray(AttributeArray&&) = default;
-    AttributeArray& operator=(AttributeArray&&) = default;
+    AttributeArray(const AttributeArray& rhs);
+    AttributeArray& operator=(const AttributeArray& rhs);
+    AttributeArray(AttributeArray&&) = delete;
+    AttributeArray& operator=(AttributeArray&&) = delete;
 
     /// Return a copy of this attribute.
     virtual AttributeArray::Ptr copy() const = 0;
 
-    /// Return an uncompressed copy of this attribute (will return a copy if not compressed).
+    /// Return a copy of this attribute.
+#ifndef _MSC_VER
+    OPENVDB_DEPRECATED_MESSAGE("In-memory compression no longer supported, use AttributeArray::copy() instead")
+#endif
     virtual AttributeArray::Ptr copyUncompressed() const = 0;
 
     /// Return the number of elements in this array.
@@ -205,7 +162,6 @@ public:
     /// @note This counts each data element in a strided array
     virtual Index dataSize() const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return the name of the value type of a single element in this array (e.g., "float" or "vec3d").
     virtual Name valueType() const = 0;
 
@@ -234,17 +190,22 @@ public:
 
     /// Return @c true if the value type is a matrix
     virtual bool valueTypeIsMatrix() const = 0;
-#endif
 
     /// Return the number of bytes of memory used by this attribute.
     virtual size_t memUsage() const = 0;
 
     /// Create a new attribute array of the given (registered) type, length and stride.
-    static Ptr create(const NamePair& type, Index length, Index stride = 1, bool constantStride = true);
+    /// @details If @a lock is non-null, the AttributeArray registry mutex
+    /// has already been locked
+    static Ptr create(const NamePair& type, Index length, Index stride = 1,
+        bool constantStride = true,
+        const Metadata* metadata = nullptr,
+        const ScopedRegistryLock* lock = nullptr);
+
     /// Return @c true if the given attribute type name is registered.
-    static bool isRegistered(const NamePair& type);
+    static bool isRegistered(const NamePair& type, const ScopedRegistryLock* lock = nullptr);
     /// Clear the attribute type registry.
-    static void clearRegistry();
+    static void clearRegistry(const ScopedRegistryLock* lock = nullptr);
 
     /// Return the name of this attribute's type.
     virtual const NamePair& type() const = 0;
@@ -257,13 +218,12 @@ public:
     bool hasValueType() const { return this->type().first == typeNameAsString<ValueType>(); }
 
     /// @brief Set value at given index @a n from @a sourceIndex of another @a sourceArray.
-    /// @deprecated From ABI 6 on, use copyValues() with source-target index pairs.
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
-    OPENVDB_DEPRECATED
+    // Windows does not allow base classes to be easily deprecated.
+#ifndef _MSC_VER
+    OPENVDB_DEPRECATED_MESSAGE("Use copyValues() with source-target index pairs")
 #endif
     virtual void set(const Index n, const AttributeArray& sourceArray, const Index sourceIndex) = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// @brief Copy values into this array from a source array to a target array
     /// as referenced by an iterator.
     /// @details Iterators must adhere to the ForwardIterator interface described
@@ -285,12 +245,15 @@ public:
     /// and both value types are floating-point or both integer.
     /// @note It is possible to use this method to write to a uniform target array
     /// if the iterator does not have non-zero target indices.
+    /// @note This method is not thread-safe, it must be guaranteed that this array is not
+    /// concurrently modified by another thread and that the source array is also not modified.
     template<typename IterT>
     void copyValuesUnsafe(const AttributeArray& sourceArray, const IterT& iter);
     /// @brief Like copyValuesUnsafe(), but if @a compact is true, attempt to collapse this array.
+    /// @note This method is not thread-safe, it must be guaranteed that this array is not
+    /// concurrently modified by another thread and that the source array is also not modified.
     template<typename IterT>
     void copyValues(const AttributeArray& sourceArray, const IterT& iter, bool compact = true);
-#endif
 
     /// Return @c true if this array is stored as a single uniform value.
     virtual bool isUniform() const = 0;
@@ -302,13 +265,16 @@ public:
     /// Compact the existing array to become uniform if all values are identical
     virtual bool compact() = 0;
 
-    /// @deprecated Previously this returned @c true if the array was compressed,
-    /// now it always returns @c false.
-    OPENVDB_DEPRECATED bool isCompressed() const { return false; }
-    /// @deprecated Previously this compressed the attribute array, now it does nothing.
-    OPENVDB_DEPRECATED virtual bool compress() = 0;
-    /// @deprecated Previously this uncompressed the attribute array, now it does nothing.
-    OPENVDB_DEPRECATED virtual bool decompress() = 0;
+    // Windows does not allow base classes to be deprecated
+#ifndef _MSC_VER
+    OPENVDB_DEPRECATED_MESSAGE("Previously this compressed the attribute array, now it does nothing")
+#endif
+    virtual bool compress() = 0;
+    // Windows does not allow base classes to be deprecated
+#ifndef _MSC_VER
+    OPENVDB_DEPRECATED_MESSAGE("Previously this uncompressed the attribute array, now it does nothing")
+#endif
+    virtual bool decompress() = 0;
 
     /// @brief   Specify whether this attribute should be hidden (e.g., from UI or iterators).
     /// @details This is useful if the attribute is used for blind data or as scratch space
@@ -369,10 +335,8 @@ public:
     /// Ensures all data is in-core
     virtual void loadData() const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return @c true if all data has been loaded
     virtual bool isDataLoaded() const = 0;
-#endif
 
     /// Check the compressed bytes and flags. If they are equal, perform a deeper
     /// comparison check necessary on the inherited types (TypedAttributeArray)
@@ -387,7 +351,6 @@ private:
     /// comparisons on inherited types
     virtual bool isEqual(const AttributeArray& other) const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Virtual function to retrieve the data buffer cast to a char byte array
     virtual char* dataAsByteArray() = 0;
     virtual const char* dataAsByteArray() const = 0;
@@ -396,9 +359,12 @@ private:
     template <typename IterT>
     void doCopyValues(const AttributeArray& sourceArray, const IterT& iter,
         bool rangeChecking = true);
-#endif
 
 protected:
+#if OPENVDB_ABI_VERSION_NUMBER >= 7
+    AttributeArray(const AttributeArray& rhs, const tbb::spin_mutex::scoped_lock&);
+#endif
+
     /// @brief Specify whether this attribute has a constant stride or not.
     void setConstantStride(bool state);
 
@@ -406,34 +372,22 @@ protected:
     virtual AccessorBasePtr getAccessor() const = 0;
 
     /// Register a attribute type along with a factory function.
-    static void registerType(const NamePair& type, FactoryMethod);
+    static void registerType(const NamePair& type, FactoryMethod,
+        const ScopedRegistryLock* lock = nullptr);
     /// Remove a attribute type from the registry.
-    static void unregisterType(const NamePair& type);
-
-#if OPENVDB_ABI_VERSION_NUMBER < 6
-
-    size_t mCompressedBytes = 0;
-    uint8_t mFlags = 0;
-    uint8_t mSerializationFlags = 0;
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
-    tbb::atomic<Index32> mOutOfCore = 0; // interpreted as bool
-#endif
-    compression::PageHandle::Ptr mPageHandle;
-
-#else // #if OPENVDB_ABI_VERSION_NUMBER < 6
+    static void unregisterType(const NamePair& type,
+        const ScopedRegistryLock* lock = nullptr);
 
     bool mIsUniform = true;
-    tbb::spin_mutex mMutex;
+    mutable tbb::spin_mutex mMutex;
     uint8_t mFlags = 0;
-    uint8_t mSerializationFlags = 0;
-    tbb::atomic<Index32> mOutOfCore = 0; // interpreted as bool
+    uint8_t mUsePagedRead = 0;
+    std::atomic<Index32> mOutOfCore; // interpreted as bool
     /// used for out-of-core, paged reading
     union {
         compression::PageHandle::Ptr mPageHandle;
-        size_t mCompressedBytes; // as of ABI=6, this data is packed together to save memory
+        size_t mCompressedBytes;
     };
-
-#endif
 }; // class AttributeArray
 
 
@@ -468,7 +422,7 @@ struct AttributeArray::Accessor : public AttributeArray::AccessorBase
 namespace attribute_traits
 {
     template <typename T> struct TruncateTrait { };
-    template <> struct TruncateTrait<float> { using Type = half; };
+    template <> struct TruncateTrait<float> { using Type = math::half; };
     template <> struct TruncateTrait<int> { using Type = short; };
 
     template <typename T> struct TruncateTrait<math::Vec3<T>> {
@@ -570,11 +524,7 @@ struct UnitVecCodec
 /// Typed class for storing attribute data
 
 template<typename ValueType_, typename Codec_ = NullCodec>
-#if OPENVDB_ABI_VERSION_NUMBER >= 6 // for ABI=6, class is final to allow for de-virtualization
 class TypedAttributeArray final: public AttributeArray
-#else
-class TypedAttributeArray: public AttributeArray
-#endif
 {
 public:
     using Ptr           = std::shared_ptr<TypedAttributeArray>;
@@ -589,25 +539,45 @@ public:
     /// Default constructor, always constructs a uniform attribute.
     explicit TypedAttributeArray(Index n = 1, Index strideOrTotalSize = 1, bool constantStride = true,
         const ValueType& uniformValue = zeroVal<ValueType>());
-    /// Deep copy constructor (optionally decompress during copy).
+#if OPENVDB_ABI_VERSION_NUMBER >= 7
+    /// Deep copy constructor.
+    /// @note This method is thread-safe (as of ABI=7) for concurrently reading from the
+    /// source attribute array while being deep-copied. Specifically, this means that the
+    /// attribute array being deep-copied can be out-of-core and safely loaded in one thread
+    /// while being copied using this copy-constructor in another thread.
+    /// It is not thread-safe for write.
+    TypedAttributeArray(const TypedAttributeArray&);
+    /// Deep copy constructor.
+    OPENVDB_DEPRECATED_MESSAGE("Use copy-constructor without unused bool parameter")
+    TypedAttributeArray(const TypedAttributeArray&, bool /*unused*/);
+#else
+    /// Deep copy constructor.
+    /// @note This method is not thread-safe for reading or writing, use
+    /// TypedAttributeArray::copy() to ensure thread-safety when reading concurrently.
     TypedAttributeArray(const TypedAttributeArray&, bool uncompress = false);
+#endif
     /// Deep copy assignment operator.
+    /// @note this operator is thread-safe.
     TypedAttributeArray& operator=(const TypedAttributeArray&);
     /// Move constructor disabled.
     TypedAttributeArray(TypedAttributeArray&&) = delete;
     /// Move assignment operator disabled.
     TypedAttributeArray& operator=(TypedAttributeArray&&) = delete;
 
-    virtual ~TypedAttributeArray() { this->deallocate(); }
+    ~TypedAttributeArray() override { this->deallocate(); }
 
     /// Return a copy of this attribute.
+    /// @note This method is thread-safe.
     AttributeArray::Ptr copy() const override;
 
-    /// Return an uncompressed copy of this attribute (will just return a copy if not compressed).
+    /// Return a copy of this attribute.
+    /// @note This method is thread-safe.
+    OPENVDB_DEPRECATED_MESSAGE("In-memory compression no longer supported, use AttributeArray::copy() instead")
     AttributeArray::Ptr copyUncompressed() const override;
 
     /// Return a new attribute array of the given length @a n and @a stride with uniform value zero.
-    static Ptr create(Index n, Index strideOrTotalSize = 1, bool constantStride = true);
+    static Ptr create(Index n, Index strideOrTotalSize = 1, bool constantStride = true,
+        const Metadata* metadata = nullptr);
 
     /// Cast an AttributeArray to TypedAttributeArray<T>
     static TypedAttributeArray& cast(AttributeArray& attributeArray);
@@ -639,7 +609,6 @@ public:
         return hasConstantStride() ? mSize * mStrideOrTotalSize : mStrideOrTotalSize;
     }
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return the name of the value type of a single element in this array (e.g., "float" or "vec3d").
     Name valueType() const override { return typeNameAsString<ValueType>(); }
 
@@ -667,38 +636,38 @@ public:
 
     /// Return @c true if the value type is a matrix
     bool valueTypeIsMatrix() const override;
-#endif
 
     /// Return the number of bytes of memory used by this attribute.
     size_t memUsage() const override;
 
-    /// Return the value at index @a n (assumes uncompressed and in-core)
+    /// Return the value at index @a n (assumes in-core)
     ValueType getUnsafe(Index n) const;
     /// Return the value at index @a n
     ValueType get(Index n) const;
-    /// Return the @a value at index @a n (assumes uncompressed and in-core)
+    /// Return the @a value at index @a n (assumes in-core)
     template<typename T> void getUnsafe(Index n, T& value) const;
     /// Return the @a value at index @a n
     template<typename T> void get(Index n, T& value) const;
 
     /// Non-member equivalent to getUnsafe() that static_casts array to this TypedAttributeArray
-    /// (assumes uncompressed and in-core)
+    /// (assumes in-core)
     static ValueType getUnsafe(const AttributeArray* array, const Index n);
 
-    /// Set @a value at the given index @a n (assumes uncompressed and in-core)
+    /// Set @a value at the given index @a n (assumes in-core)
     void setUnsafe(Index n, const ValueType& value);
     /// Set @a value at the given index @a n
     void set(Index n, const ValueType& value);
-    /// Set @a value at the given index @a n (assumes uncompressed and in-core)
+    /// Set @a value at the given index @a n (assumes in-core)
     template<typename T> void setUnsafe(Index n, const T& value);
     /// Set @a value at the given index @a n
     template<typename T> void set(Index n, const T& value);
 
     /// Non-member equivalent to setUnsafe() that static_casts array to this TypedAttributeArray
-    /// (assumes uncompressed and in-core)
+    /// (assumes in-core)
     static void setUnsafe(AttributeArray* array, const Index n, const ValueType& value);
 
     /// Set value at given index @a n from @a sourceIndex of another @a sourceArray
+    OPENVDB_DEPRECATED_MESSAGE("Use copyValues() with source-target index pairs")
     void set(const Index n, const AttributeArray& sourceArray, const Index sourceIndex) override;
 
     /// Return @c true if this array is stored as a single uniform value.
@@ -724,9 +693,11 @@ public:
     static void fill(AttributeArray* array, const ValueType& value);
 
     /// Compress the attribute array.
-    OPENVDB_DEPRECATED bool compress() override;
+    OPENVDB_DEPRECATED_MESSAGE("Previously this compressed the attribute array, now it does nothing")
+    bool compress() override;
     /// Uncompress the attribute array.
-    OPENVDB_DEPRECATED bool decompress() override;
+    OPENVDB_DEPRECATED_MESSAGE("Previously this uncompressed the attribute array, now it does nothing")
+    bool decompress() override;
 
     /// Read attribute data from a stream.
     void read(std::istream&) override;
@@ -765,10 +736,8 @@ public:
     /// Ensures all data is in-core
     void loadData() const override;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return @c true if all data has been loaded
     bool isDataLoaded() const override;
-#endif
 
 protected:
     AccessorBasePtr getAccessor() const override;
@@ -783,10 +752,14 @@ protected:
 private:
     friend class ::TestAttributeArray;
 
+#if OPENVDB_ABI_VERSION_NUMBER >= 7
+    TypedAttributeArray(const TypedAttributeArray&, const tbb::spin_mutex::scoped_lock&);
+#endif
+
     /// Load data from memory-mapped file.
     inline void doLoad() const;
     /// Load data from memory-mapped file (unsafe as this function is not protected by a mutex).
-    /// @param compression if true, loading previously compressed data will re-compressed it
+    /// @param compression parameter no longer used
     inline void doLoadUnsafe(const bool compression = true) const;
     /// Compress in-core data assuming mutex is locked
     inline bool compressUnsafe();
@@ -797,29 +770,24 @@ private:
     /// Compare the this data to another attribute array. Used by the base class comparison operator
     bool isEqual(const AttributeArray& other) const override;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Virtual function to retrieve the data buffer from the derived class cast to a char byte array
     char* dataAsByteArray() override;
     const char* dataAsByteArray() const override;
-#endif
 
     size_t arrayMemUsage() const;
     void allocate();
     void deallocate();
 
     /// Helper function for use with registerType()
-    static AttributeArray::Ptr factory(Index n, Index strideOrTotalSize, bool constantStride) {
-        return TypedAttributeArray::create(n, strideOrTotalSize, constantStride);
+    static AttributeArray::Ptr factory(Index n, Index strideOrTotalSize, bool constantStride,
+        const Metadata* metadata) {
+        return TypedAttributeArray::create(n, strideOrTotalSize, constantStride, metadata);
     }
 
-    static tbb::atomic<const NamePair*> sTypeName;
+    static std::unique_ptr<const NamePair> sTypeName;
     std::unique_ptr<StorageType[]>      mData;
     Index                               mSize;
     Index                               mStrideOrTotalSize;
-#if OPENVDB_ABI_VERSION_NUMBER < 6 // as of ABI=6, this data lives in the base class to reduce memory
-    bool                                mIsUniform = true;
-    tbb::spin_mutex                     mMutex;
-#endif
 }; // class TypedAttributeArray
 
 
@@ -1028,8 +996,6 @@ UnitVecCodec::encode(const math::Vec3<T>& val, StorageType& data)
 
 // AttributeArray implementation
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
-
 template <typename IterT>
 void AttributeArray::doCopyValues(const AttributeArray& sourceArray, const IterT& iter,
     bool rangeChecking/*=true*/)
@@ -1105,6 +1071,12 @@ void AttributeArray::copyValues(const AttributeArray& sourceArray, const IterT&
     // if the target array is uniform, expand it first
     this->expand();
 
+    // TODO: Acquire mutex locks for source and target arrays to ensure that
+    // value copying is always thread-safe. Note that the unsafe method will be
+    // faster, but can only be used if neither the source or target arrays are
+    // modified during copying. Note that this will require a new private
+    // virtual method with ABI=7 to access the mutex from the derived class.
+
     this->doCopyValues(sourceArray, iter, true);
 
     // attempt to compact target array
@@ -1112,7 +1084,6 @@ void AttributeArray::copyValues(const AttributeArray& sourceArray, const IterT&
         this->compact();
     }
 }
-#endif
 
 
 ////////////////////////////////////////
@@ -1120,7 +1091,7 @@ void AttributeArray::copyValues(const AttributeArray& sourceArray, const IterT&
 // TypedAttributeArray implementation
 
 template<typename ValueType_, typename Codec_>
-tbb::atomic<const NamePair*> TypedAttributeArray<ValueType_, Codec_>::sTypeName;
+std::unique_ptr<const NamePair> TypedAttributeArray<ValueType_, Codec_>::sTypeName;
 
 
 template<typename ValueType_, typename Codec_>
@@ -1151,18 +1122,29 @@ TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(
 }
 
 
+#if OPENVDB_ABI_VERSION_NUMBER >= 7
+template<typename ValueType_, typename Codec_>
+TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs)
+    : TypedAttributeArray(rhs, tbb::spin_mutex::scoped_lock(rhs.mMutex))
+{
+}
+
+
+template<typename ValueType_, typename Codec_>
+TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs,
+    const tbb::spin_mutex::scoped_lock& lock)
+    : AttributeArray(rhs, lock)
+#else
 template<typename ValueType_, typename Codec_>
 TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs, bool)
     : AttributeArray(rhs)
+#endif
     , mSize(rhs.mSize)
     , mStrideOrTotalSize(rhs.mStrideOrTotalSize)
-#if OPENVDB_ABI_VERSION_NUMBER < 6
-    , mIsUniform(rhs.mIsUniform)
-#endif
 {
     if (this->validData()) {
         this->allocate();
-        std::memcpy(this->data(), rhs.data(), this->arrayMemUsage());
+        std::memcpy(static_cast<void*>(this->data()), rhs.data(), this->arrayMemUsage());
     }
 }
 
@@ -1172,21 +1154,25 @@ TypedAttributeArray<ValueType_, Codec_>&
 TypedAttributeArray<ValueType_, Codec_>::operator=(const TypedAttributeArray& rhs)
 {
     if (&rhs != this) {
+        // lock both the source and target arrays to ensure thread-safety
         tbb::spin_mutex::scoped_lock lock(mMutex);
+        tbb::spin_mutex::scoped_lock rhsLock(rhs.mMutex);
 
         this->deallocate();
 
         mFlags = rhs.mFlags;
-        mSerializationFlags = rhs.mSerializationFlags;
+        mUsePagedRead = rhs.mUsePagedRead;
         mSize = rhs.mSize;
         mStrideOrTotalSize = rhs.mStrideOrTotalSize;
         mIsUniform = rhs.mIsUniform;
 
         if (this->validData()) {
             this->allocate();
-            std::memcpy(this->newDataAsByteArray(), rhs.newDataAsByteArray(), this->arrayMemUsage());
+            std::memcpy(static_cast<void*>(this->data()), rhs.data(), this->arrayMemUsage());
         }
     }
+
+    return *this;
 }
 
 
@@ -1194,10 +1180,11 @@ template<typename ValueType_, typename Codec_>
 inline const NamePair&
 TypedAttributeArray<ValueType_, Codec_>::attributeType()
 {
-    if (sTypeName == nullptr) {
-        NamePair* s = new NamePair(typeNameAsString<ValueType>(), Codec::name());
-        if (sTypeName.compare_and_swap(s, nullptr) != nullptr) delete s;
-    }
+    static std::once_flag once;
+    std::call_once(once, []()
+    {
+        sTypeName.reset(new NamePair(typeNameAsString<ValueType>(), Codec::name()));
+    });
     return *sTypeName;
 }
 
@@ -1228,9 +1215,14 @@ TypedAttributeArray<ValueType_, Codec_>::unregisterType()
 
 template<typename ValueType_, typename Codec_>
 inline typename TypedAttributeArray<ValueType_, Codec_>::Ptr
-TypedAttributeArray<ValueType_, Codec_>::create(Index n, Index stride, bool constantStride)
+TypedAttributeArray<ValueType_, Codec_>::create(Index n, Index stride, bool constantStride,
+    const Metadata* metadata)
 {
-    return Ptr(new TypedAttributeArray(n, stride, constantStride));
+    const TypedMetadata<ValueType>* typedMetadata = metadata ?
+        dynamic_cast<const TypedMetadata<ValueType>*>(metadata) : nullptr;
+
+    return Ptr(new TypedAttributeArray(n, stride, constantStride,
+        typedMetadata ? typedMetadata->value() : zeroVal<ValueType>()));
 }
 
 template<typename ValueType_, typename Codec_>
@@ -1257,6 +1249,9 @@ template<typename ValueType_, typename Codec_>
 AttributeArray::Ptr
 TypedAttributeArray<ValueType_, Codec_>::copy() const
 {
+#if OPENVDB_ABI_VERSION_NUMBER < 7
+    tbb::spin_mutex::scoped_lock lock(mMutex);
+#endif
     return AttributeArray::Ptr(new TypedAttributeArray<ValueType, Codec>(*this));
 }
 
@@ -1265,7 +1260,7 @@ template<typename ValueType_, typename Codec_>
 AttributeArray::Ptr
 TypedAttributeArray<ValueType_, Codec_>::copyUncompressed() const
 {
-    return AttributeArray::Ptr(new TypedAttributeArray<ValueType, Codec>(*this, /*decompress = */true));
+    return this->copy();
 }
 
 
@@ -1308,7 +1303,6 @@ TypedAttributeArray<ValueType_, Codec_>::deallocate()
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::valueTypeIsFloatingPoint() const
@@ -1325,7 +1319,7 @@ TypedAttributeArray<ValueType_, Codec_>::valueTypeIsFloatingPoint() const
     using ElementT = typename VecTraits<ValueType>::ElementType;
 
     // half is not defined as float point as expected, so explicitly handle it
-    return std::is_floating_point<ElementT>::value || std::is_same<half, ElementT>::value;
+    return std::is_floating_point<ElementT>::value || std::is_same<math::half, ElementT>::value;
 }
 
 
@@ -1334,7 +1328,7 @@ bool
 TypedAttributeArray<ValueType_, Codec_>::valueTypeIsClass() const
 {
     // half is not defined as a non-class type as expected, so explicitly exclude it
-    return std::is_class<ValueType>::value && !std::is_same<half, ValueType>::value;
+    return std::is_class<ValueType>::value && !std::is_same<math::half, ValueType>::value;
 }
 
 
@@ -1362,7 +1356,6 @@ TypedAttributeArray<ValueType_, Codec_>::valueTypeIsMatrix() const
     // TODO: improve performance by making this a compile-time check using type traits
     return !this->valueType().compare(0, 3, "mat");
 }
-#endif
 
 
 template<typename ValueType_, typename Codec_>
@@ -1609,11 +1602,7 @@ template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::isOutOfCore() const
 {
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     return mOutOfCore;
-#else
-    return (mFlags & OUTOFCORE);
-#endif
 }
 
 
@@ -1621,12 +1610,7 @@ template<typename ValueType_, typename Codec_>
 void
 TypedAttributeArray<ValueType_, Codec_>::setOutOfCore(const bool b)
 {
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     mOutOfCore = b;
-#else
-    if (b) mFlags = static_cast<uint8_t>(mFlags | OUTOFCORE);
-    else   mFlags = static_cast<uint8_t>(mFlags & ~OUTOFCORE);
-#endif
 }
 
 
@@ -1654,14 +1638,12 @@ TypedAttributeArray<ValueType_, Codec_>::loadData() const
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::isDataLoaded() const
 {
     return !this->isOutOfCore();
 }
-#endif
 
 
 template<typename ValueType_, typename Codec_>
@@ -1689,7 +1671,6 @@ TypedAttributeArray<ValueType_, Codec_>::readMetadata(std::istream& is)
 
     uint8_t serializationFlags = uint8_t(0);
     is.read(reinterpret_cast<char*>(&serializationFlags), sizeof(uint8_t));
-    mSerializationFlags = serializationFlags;
 
     Index size = Index(0);
     is.read(reinterpret_cast<char*>(&size), sizeof(Index));
@@ -1701,19 +1682,20 @@ TypedAttributeArray<ValueType_, Codec_>::readMetadata(std::istream& is)
     }
     // error if an unknown serialization flag has been set,
     // as this will adjust the layout of the data and corrupt the ability to read
-    if (mSerializationFlags >= 0x10) {
+    if (serializationFlags >= 0x10) {
         OPENVDB_THROW(IoError, "Unknown attribute serialization flags for VDB file format.");
     }
 
-    // read uniform and compressed state
+    // set uniform, compressed and page read state
 
-    mIsUniform = mSerializationFlags & WRITEUNIFORM;
+    mIsUniform = serializationFlags & WRITEUNIFORM;
+    mUsePagedRead = serializationFlags & WRITEPAGED;
     mCompressedBytes = bytes;
     mFlags |= PARTIALREAD; // mark data as having been partially read
 
     // read strided value (set to 1 if array is not strided)
 
-    if (mSerializationFlags & WRITESTRIDED) {
+    if (serializationFlags & WRITESTRIDED) {
         Index stride = Index(0);
         is.read(reinterpret_cast<char*>(&stride), sizeof(Index));
         mStrideOrTotalSize = stride;
@@ -1728,7 +1710,7 @@ template<typename ValueType_, typename Codec_>
 void
 TypedAttributeArray<ValueType_, Codec_>::readBuffers(std::istream& is)
 {
-    if ((mSerializationFlags & WRITEPAGED)) {
+    if (mUsePagedRead) {
         // use readBuffers(PagedInputStream&) for paged buffers
         OPENVDB_THROW(IoError, "Cannot read paged AttributeArray buffers.");
     }
@@ -1760,11 +1742,6 @@ TypedAttributeArray<ValueType_, Codec_>::readBuffers(std::istream& is)
     // set data to buffer
 
     mData.reset(reinterpret_cast<StorageType*>(buffer.release()));
-
-    // clear all write flags
-
-    if (mIsUniform)     mSerializationFlags &= uint8_t(~WRITEUNIFORM & ~WRITEMEMCOMPRESS & ~WRITEPAGED);
-    else                mSerializationFlags &= uint8_t(~WRITEUNIFORM & ~WRITEPAGED);
 }
 
 
@@ -1772,7 +1749,7 @@ template<typename ValueType_, typename Codec_>
 void
 TypedAttributeArray<ValueType_, Codec_>::readPagedBuffers(compression::PagedInputStream& is)
 {
-    if (!(mSerializationFlags & WRITEPAGED)) {
+    if (!mUsePagedRead) {
         if (!is.sizeOnly()) this->readBuffers(is.getInputStream());
         return;
     }
@@ -1804,12 +1781,12 @@ TypedAttributeArray<ValueType_, Codec_>::readPagedBuffers(compression::PagedInpu
     if (!delayLoad) {
         std::unique_ptr<char[]> buffer = mPageHandle->read();
         mData.reset(reinterpret_cast<StorageType*>(buffer.release()));
+        mPageHandle.reset();
     }
 
-    // clear all write flags
+    // clear page state
 
-    if (mIsUniform)     mSerializationFlags &= uint8_t(~WRITEUNIFORM & ~WRITEMEMCOMPRESS & ~WRITEPAGED);
-    else                mSerializationFlags &= uint8_t(~WRITEUNIFORM & ~WRITEPAGED);
+    mUsePagedRead = 0;
 }
 
 
@@ -1840,11 +1817,7 @@ TypedAttributeArray<ValueType_, Codec_>::writeMetadata(std::ostream& os, bool ou
         OPENVDB_THROW(IoError, "Cannot write out a partially-read AttributeArray.");
     }
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     uint8_t flags(mFlags);
-#else
-    uint8_t flags(mFlags & uint8_t(~OUTOFCORE));
-#endif
     uint8_t serializationFlags(0);
     Index size(mSize);
     Index stride(mStrideOrTotalSize);
@@ -1969,6 +1942,7 @@ TypedAttributeArray<ValueType_, Codec_>::doLoadUnsafe(const bool /*compression*/
     auto* self = const_cast<TypedAttributeArray<ValueType_, Codec_>*>(this);
 
     assert(self->mPageHandle);
+    assert(!(self->mFlags & PARTIALREAD));
 
     std::unique_ptr<char[]> buffer = self->mPageHandle->read();
 
@@ -1978,12 +1952,7 @@ TypedAttributeArray<ValueType_, Codec_>::doLoadUnsafe(const bool /*compression*/
 
     // clear all write and out-of-core flags
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     self->mOutOfCore = false;
-#else
-    self->mFlags &= uint8_t(~OUTOFCORE);
-#endif
-    self->mSerializationFlags &= uint8_t(~WRITEUNIFORM & ~WRITEMEMCOMPRESS & ~WRITEPAGED);
 }
 
 
@@ -1992,7 +1961,7 @@ AttributeArray::AccessorBasePtr
 TypedAttributeArray<ValueType_, Codec_>::getAccessor() const
 {
     // use the faster 'unsafe' get and set methods as attribute handles
-    // ensure data is uncompressed and in-core when constructed
+    // ensure data is in-core when constructed
 
     return AccessorBasePtr(new AttributeArray::Accessor<ValueType_>(
         &TypedAttributeArray<ValueType_, Codec_>::getUnsafe,
@@ -2011,7 +1980,7 @@ TypedAttributeArray<ValueType_, Codec_>::isEqual(const AttributeArray& other) co
     if(this->mSize != otherT->mSize ||
        this->mStrideOrTotalSize != otherT->mStrideOrTotalSize ||
        this->mIsUniform != otherT->mIsUniform ||
-       *this->sTypeName != *otherT->sTypeName) return false;
+       this->attributeType() != this->attributeType()) return false;
 
     this->doLoad();
     otherT->doLoad();
@@ -2025,7 +1994,6 @@ TypedAttributeArray<ValueType_, Codec_>::isEqual(const AttributeArray& other) co
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 char*
 TypedAttributeArray<ValueType_, Codec_>::dataAsByteArray()
@@ -2040,7 +2008,6 @@ TypedAttributeArray<ValueType_, Codec_>::dataAsByteArray() const
 {
     return reinterpret_cast<const char*>(this->data());
 }
-#endif
 
 
 ////////////////////////////////////////
@@ -2300,7 +2267,3 @@ AttributeArray& AttributeWriteHandle<ValueType, CodecType>::array()
 } // namespace openvdb
 
 #endif // OPENVDB_POINTS_ATTRIBUTE_ARRAY_HAS_BEEN_INCLUDED
-
-// Copyright (c) 2012-2018 DreamWorks Animation LLC
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
